import dbm
import os
import json
from telegram import InlineKeyboardButton
from .papers_args import add_arguments_papers, parse_args_papers
from .papers_display import papers_results2message

favorites_dirname = os.environ['RABBOT_FAVORITE_DIR'] if 'RABBOT_FAVORITE_DIR' in os.environ else "save/favorites"


def favorite_paper_parser_add_arguments(parser):
    parser.add_argument('key', metavar='key', type=str,
                        help='Key and value to identify the paper')
    parser.add_argument('value', metavar='value', type=str,
                        help='Key and value to identify the paper')
    return parser


def add_favorite_paper_args2querys(args: object):
    username = args.update.message.from_user.username
    favorites_paper_path = os.path.join(favorites_dirname, username, "papers")
    os.makedirs(os.path.dirname(favorites_paper_path), exist_ok=True)
    k, v = args.key, args.value.encode(encoding="utf8")
    with dbm.open(favorites_paper_path, 'c') as db:
        if v not in db:
            db[v] = json.dumps([k]).encode(encoding="utf8")
        else:
            old_k = json.loads(db[v].decode(encoding="utf8"))
            new_k = list(set([k, *old_k]))
            db[v] = json.dumps(new_k).encode(encoding="utf8")
    return [(f"MATCH (p:Publication) WHERE p.{args.key}=$value RETURN p.title", {"value": args.value})]


def rm_favorite_paper_args2querys(args: object):
    username = args.update.message.from_user.username
    favorites_paper_path = os.path.join(favorites_dirname, username, "papers")
    os.makedirs(os.path.dirname(favorites_paper_path), exist_ok=True)
    k, v = args.key, args.value.encode(encoding="utf8")
    with dbm.open(favorites_paper_path, 'c') as db:
        if v in db:
            old_k = set(json.loads(db[v].decode(encoding="utf8")))
            if k in old_k:
                old_k.remove(k)
                new_k = list(old_k)
                if len(new_k) <= 0:
                    del db[v]
                else:
                    db[v] = json.dumps(new_k).encode(encoding="utf8")
    return [(f"MATCH (p:Publication) WHERE p.{args.key}=$value RETURN p.title", {"value": args.value})]


def add_favorite_paper_results2message(res, args: object):
    msg = f"Paper added in favorites. {args.key}:{args.value}"
    for r in res[0]:
        msg += "\n" + f"<b>{r[0]}</b>"
    return msg, [[
        InlineKeyboardButton(
            f'Remove it',
            switch_inline_query_current_chat=f'/rm_favorite_paper "{args.key}" "{args.value}"'),
        InlineKeyboardButton(
            f'Show all',
            switch_inline_query_current_chat=f'/show_favorite_paper -o date')
    ]]


def rm_favorite_paper_results2message(res, args: object):
    msg = f"Paper removed from favorites. {args.key}:{args.value}"
    for r in res[0]:
        msg += "\n" + f"<b>{r[0]}</b>"
    return msg, [[
        InlineKeyboardButton(
            f'Revoke it',
            switch_inline_query_current_chat=f'/add_favorite_paper "{args.key}" "{args.value}"'),
        InlineKeyboardButton(
            f'Show all',
            switch_inline_query_current_chat=f'/show_favorite_paper -o date')
    ]]


add_favorite_paper_jump = (
    "add_favorite_paper",
    favorite_paper_parser_add_arguments,
    add_favorite_paper_args2querys,
    add_favorite_paper_results2message,
    None
)


rm_favorite_paper_jump = (
    "rm_favorite_paper",
    favorite_paper_parser_add_arguments,
    rm_favorite_paper_args2querys,
    rm_favorite_paper_results2message,
    None
)


def favorite_paper_args2querys(args: object):
    username = args.update.message.from_user.username
    favorites_paper_path = os.path.join(favorites_dirname, username, "papers")
    os.makedirs(os.path.dirname(favorites_paper_path), exist_ok=True)
    pwhere, jwhere, orderby, limits, values = parse_args_papers(args)
    k_or, v_or = [], {}
    with dbm.open(favorites_paper_path, 'c') as db:
        i = 0
        for v in db.keys():
            ks = db[v]
            for k in json.loads(ks.decode("utf8")):
                k_or.append(f"p.{k}=$fav{i}")
                v_or[f"fav{i}"] = v.decode("utf8")
                i += 1
    if len(k_or) <= 0:
        return []
    pwhere += " AND " + f"({' or '.join(k_or)})"
    values = {**values, **v_or}
    return [(
        f"MATCH (p:Publication) WHERE {pwhere} " +
        ("OPTIONAL MATCH (p:Publication)-[:PUBLISH]->(j:Journal) " if jwhere == '' else f"MATCH (p:Publication)-[:PUBLISH]->(j:Journal) WHERE {jwhere} ") +
        f"OPTIONAL MATCH (c:Publication)-[:CITE]->(p:Publication) "
        f"OPTIONAL MATCH (r:Publication)<-[:CITE]-(p:Publication) "
        f"RETURN p, j, COUNT(DISTINCT c) AS citation, COUNT(DISTINCT r) AS reference ORDER BY {orderby} LIMIT {limits}",
        {**values}
    )]


show_favorite_paper_jump = (
    "show_favorite_paper",
    add_arguments_papers,
    favorite_paper_args2querys,
    papers_results2message,
    ("Show my favorite papers", "-o date")
)


def add_favorite_keywords_args2querys(args: object):
    username = args.update.message.from_user.username
    favorites_keywords_path = os.path.join(favorites_dirname, username, "keywords")
    os.makedirs(os.path.dirname(favorites_keywords_path), exist_ok=True)
    with dbm.open(favorites_keywords_path, 'c') as db:
        for keyword in args.keyword:
            db[keyword.encode(encoding="utf8")] = ''
    return [(f"RETURN true", {})]


def show_favorite_keywords_results2message(_, args: object):
    username = args.update.message.from_user.username
    favorites_keywords_path = os.path.join(favorites_dirname, username, "keywords")
    os.makedirs(os.path.dirname(favorites_keywords_path), exist_ok=True)
    msg = "Your favorited keywords:"
    keyboard = []
    with dbm.open(favorites_keywords_path, 'c') as db:
        for keyword in db.keys():
            keyword = keyword.decode(encoding="utf8")
            msg += "\n" + f"<b>{keyword}</b>"
            keyboard.append([
                InlineKeyboardButton(
                    f'Use "{keyword}"',
                    switch_inline_query_current_chat=f'/search_by_title -k "{keyword}" -o date'),
                InlineKeyboardButton(
                    f'Remove "{keyword}"',
                    switch_inline_query_current_chat=f'/rm_favorite_keywords -k "{keyword}"')
            ])
    return msg, keyboard


add_favorite_keywords_jump = (
    "add_favorite_keywords",
    add_arguments_papers,
    add_favorite_keywords_args2querys,
    show_favorite_keywords_results2message,
    None
)


show_favorite_keywords_jump = (
    "show_favorite_keywords",
    None,
    lambda _: [(f"RETURN true", {})],
    show_favorite_keywords_results2message,
    ("Show my favorite keywords", "")
)


def rm_favorite_keywords_args2querys(args: object):
    username = args.update.message.from_user.username
    favorites_keywords_path = os.path.join(favorites_dirname, username, "keywords")
    os.makedirs(os.path.dirname(favorites_keywords_path), exist_ok=True)
    with dbm.open(favorites_keywords_path, 'c') as db:
        for keyword in args.keyword:
            if keyword.encode(encoding="utf8") in db:
                del db[keyword.encode(encoding="utf8")]
    return [(f"RETURN true", {})]


def rm_favorite_keywords_results2message(_, args: object):
    msg = "The following keywords removed:"
    keyboard = []
    for keyword in args.keyword:
        msg += "\n" + f"<b>{keyword}</b>"
        keyboard.append([
            InlineKeyboardButton(
                f'Revoke "{keyword}"',
                switch_inline_query_current_chat=f'/add_favorite_keywords -k "{keyword}"'),
            InlineKeyboardButton(
                f'Show all',
                switch_inline_query_current_chat=f'/show_favorite_keywords')
        ])
    return msg, keyboard


rm_favorite_keywords_jump = (
    "rm_favorite_keywords",
    add_arguments_papers,
    rm_favorite_keywords_args2querys,
    rm_favorite_keywords_results2message,
    None
)



from .papers import author_papers_jump, citations_jump, references_jump
from .search import search_by_title_jump
from .detail import paper_detail_jump
from .favorites import add_favorite_paper_jump, show_favorite_paper_jump, rm_favorite_paper_jump
from .favorites import add_favorite_keywords_jump, show_favorite_keywords_jump, rm_favorite_keywords_jump



import asyncio
import re
import shlex
from argparse import ArgumentParser
from typing import Callable, Tuple, Dict, List, Sequence, NamedTuple
from neo4j import Session
from telegram import Update, InlineKeyboardMarkup, InlineKeyboardButton
from telegram.ext import filters, Application, ContextTypes, CommandHandler, MessageHandler
from telegram.constants import ParseMode


class ObjArgs(NamedTuple):
    update: Update
    context: ContextTypes.DEFAULT_TYPE


class Rabbot:
    def __init__(self, app: Application, session: Session) -> None:
        self.app = app
        self.session = session
        self.bot_name = None

    async def _fetch_name(self):
        me = await self.app.bot.get_me()
        self.bot_name = me.name

    def add_jump(self, name: str,
                 parser_add_arguments: Callable[[ArgumentParser], ArgumentParser],
                 args2querys: Callable[[object], List[Tuple[str, Dict]]],
                 results2message: Callable[[List, object], Tuple[str, Sequence[Sequence[InlineKeyboardButton]]]]):
        parser = ArgumentParser(add_help=False, exit_on_error=False)
        parser.add_argument('-h', '--help', action='store_true', help='Show help message.')
        if parser_add_arguments:
            parser = parser_add_arguments(parser)

        async def handler(text: str, update: Update, context: ContextTypes.DEFAULT_TYPE):
            lst_args = []
            str_args = re.findall(r"^/[A-Za-z_]+ *(.*)$", text)
            obj_args = None
            if len(str_args) > 0:
                lst_args = shlex.split(str_args[0])
            if parser_add_arguments is not None:
                if len(lst_args) <= 0 or '-h' in lst_args or '--help' in lst_args:
                    message = re.sub(r"^usage: __main__.py", f"usage: /{name}", parser.format_help())
                    await update.message.reply_text(text=message, reply_to_message_id=update.message.id)
                    return
                else:
                    obj_args = parser.parse_args(lst_args)
                    if not hasattr(obj_args, "update"):
                        setattr(obj_args, "update", update)
                    if not hasattr(obj_args, "context"):
                        setattr(obj_args, "context", context)
            else:
                obj_args = ObjArgs(update=update, context=context)
            querys = args2querys(obj_args)
            if not querys or len(querys) <= 0:
                return
            results = []
            for query, kwargs in querys:
                results.append(self.session.execute_read(lambda tx: tx.run(query, **kwargs).values()))
            message, keyboard = results2message(results, obj_args)
            reply_markup = InlineKeyboardMarkup([
                *keyboard, [InlineKeyboardButton(
                    "Try again",
                    switch_inline_query_current_chat=f"/{name} {str_args[0] if len(str_args) > 0 else ''}"
                )]
            ])
            await update.message.reply_text(text=message, reply_to_message_id=update.message.id, reply_markup=reply_markup, parse_mode=ParseMode.HTML)
        command_handler = CommandHandler(name, lambda update, context: handler(update.message.text, update, context))
        self.app.add_handler(command_handler)

        if self.bot_name is None:
            loop = asyncio.get_event_loop()
            loop.run_until_complete(self._fetch_name())

        async def wrap_handler(update: Update, context: ContextTypes.DEFAULT_TYPE):
            split = re.findall(f"^{self.bot_name} +(/{name} *.*)", update.message.text)
            if len(split) > 0:
                await handler(split[0], update, context)
        message_handler = MessageHandler(filters.Regex(f"^{self.bot_name} +/{name}"), wrap_handler)
        self.app.add_handler(message_handler)



from telegram import InlineKeyboardButton
from typing import List


def reconstruct_paper_args(args: object):
    paper_args = ""
    paper_args += f"--order {args.order} "
    paper_args += f"--year {args.year} "
    paper_args += ' '.join([f'--keyword "{k}"' for k in args.keyword]) + " "
    paper_args += ' '.join([f'--where_paper "{kv}"' for kv in args.where_paper]) + " "
    paper_args += ' '.join([f'--where_journal "{kv}"' for kv in args.where_journal]) + " "
    paper_args += f"--limit {args.limit}"
    return paper_args


def papers_results2message(res: List, args: object):
    msg = ""
    paper_args = reconstruct_paper_args(args)
    keyboards = []
    for i, (paper, journal, n_cite, n_refs) in enumerate(res[0]):
        title = paper['title']
        journal_info = "not published"
        if journal:
            journal_info = f"<i>{journal['dblp_name']}</i> (CCF {journal['ccf']})"
        msg += f'<b>P{i+1}.</b> '
        if "paperId" in paper:
            msg += f'<a href="https://www.semanticscholar.org/paper/{paper["paperId"]}"><b>{title}</b></a>'
        elif "doi" in paper:
            msg += f'<a href="https://doi.org/{paper["doi"]}"><b>{title}</b></a>'
        else:
            msg += f"<b>{title}</b>"
        msg += f", {journal_info}, {paper['date'] if 'date' in paper else paper['year']}"
        msg += f", {n_refs} references, <b>{n_cite} citations</b>\n"

        keyboards.append(
            InlineKeyboardButton(
                f"P{i+1} Detail",
                switch_inline_query_current_chat=f"/paper_detail {paper_args} title_hash {paper['title_hash']}")
        )
    N = 3
    keyboard = []
    for i in range(len(keyboards) // N + 1):
        keyboard.append([])
        for j in range(N):
            if i*3+j < len(keyboards):
                keyboard[-1].append(keyboards[i*3+j])
    if msg == "":
        msg = "No papers yet"
    keyboard.append([
        InlineKeyboardButton(
            f"Add Favorite Keywords",
            switch_inline_query_current_chat=f'/add_favorite_keywords {paper_args}'),
        InlineKeyboardButton(
            f"Remove Favorite Keywords",
            switch_inline_query_current_chat=f'/rm_favorite_keywords {paper_args}')
    ])
    return msg, keyboard
